<?php

namespace Drupal\xnttdb\Plugin\ExternalEntities\StorageClient;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\Query\Sql\Condition;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\dbxschema\Database\DatabaseTool;
use Drupal\external_entities\ExternalEntityInterface;
use Drupal\external_entities\ExternalEntityTypeInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface;
use Drupal\external_entities\StorageClient\ExternalEntityStorageClientBase;
use Drupal\xnttdb\Exception\DatabaseExternalEntityException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * External entities storage client for external databases/schemas.
 *
 * @ExternalEntityStorageClient(
 *   id = "xnttdb",
 *   label = @Translation("Database"),
 *   description = @Translation("Retrieves database records.")
 * )
 */
class Database extends ExternalEntityStorageClientBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * Default schema name identifier.
   */
  const DEFAULT_SCHEMA = '<default>';

  /**
   * Available cross connections.
   *
   * @var array
   */
  protected static $xConnections;

  /**
   * The connection to use.
   *
   * @var \Drupal\dbxschema\Database\CrossSchemaConnectionInterface
   */
  protected $xConnection;

  /**
   * Placeholder data.
   *
   * Keys are placeholder names with schema prefix and values are replacement
   * values.
   *
   * @var array
   */
  protected $placeholders;

  /**
   * Entity field manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * Cache backend service.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * Logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Database tool service.
   *
   * @var \Drupal\dbxschema\Database\DatabaseTool
   */
  protected $databaseTool;

  /**
   * Constructs a database external storage object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\external_entities\ResponseDecoder\ResponseDecoderFactoryInterface $response_decoder_factory
   *   The response decoder factory service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger service.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   Cache backend service.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   Logger service (xnttdb).
   * @param \Drupal\dbxschema\Database\DatabaseTool $database_tool
   *   Database tool service.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    ResponseDecoderFactoryInterface $response_decoder_factory,
    EntityFieldManagerInterface $entity_field_manager,
    MessengerInterface $messenger,
    CacheBackendInterface $cache,
    LoggerChannelInterface $logger,
    DatabaseTool $database_tool
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $string_translation,
      $response_decoder_factory
    );

    // Services injection.
    $this->entityFieldManager = $entity_field_manager;
    $this->messenger = $messenger;
    $this->cache = $cache;
    $this->logger = $logger;
    $this->databaseTool = $database_tool;

    // Gets schema and default schema.
    $schemas = $configuration['connection']['schemas'] ?? [];
    $default_schema = array_shift($schemas);
    // Get database key if one.
    $dbkey = $configuration['connection']['dbkey'];
    if (!empty($default_schema)) {
      $ckey = $dbkey . '#' . $default_schema . '#' . implode('#', $schemas);
      // Try to get connection from static connection pool.
      if (!empty(static::$xConnections[$ckey])) {
        $this->xConnection = static::$xConnections[$ckey];
        try {
          // Initializes placeholders replacement.
          $this->initPlaceholders();
        }
        catch (DatabaseExternalEntityException $e) {
          $this->messenger->addError(
            'Database External Entity initialization failed: ' . $e->getMessage()
          );
        }
      }
      else {
        // Initializes connection, first schema and extra-schemas.
        try {
          if (empty($default_schema)
              || (strtolower($default_schema) == static::DEFAULT_SCHEMA)
          ) {
            $default_schema = '';
          }
          // Get an external schema connection.
          $connection = $this->databaseTool->getConnection($default_schema, $dbkey);
          // Initialize connection.
          if (!$connection->schema()->schemaExists()) {
            throw new DatabaseExternalEntityException(
              'Invalid schema "'
              . $default_schema
              . '". The schema does not exist.'
            );
          }
          // Get related database tool (instead of default provided one).
          $db_tool = $this->databaseTool = $connection->getDatabaseTool();
          // Adds extra schemas.
          foreach ($schemas as $schema_name) {
            if (!$db_tool->schemaExists($schema_name)) {
              throw new DatabaseExternalEntityException(
                'Invalid schema "'
                . $schema_name
                . '". The schema does not exist.'
              );
            }
            $connection->addExtraSchema($schema_name);
          }
          static::$xConnections[$ckey] = $this->xConnection = $connection;

          // Initializes placeholders replacement.
          $this->initPlaceholders();
        }
        catch (DatabaseExternalEntityException $e) {
          $this->messenger->addError(
            'Database External Entity initialization failed: ' . $e->getMessage()
          );
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('external_entities.response_decoder_factory'),
      $container->get('entity_field.manager'),
      $container->get('messenger'),
      $container->get('cache.default'),
      $container->get('logger.channel.xnttdb'),
      $container->get('dbxschema.tool')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'qcreate' => NULL,
      'qread'   => NULL,
      'qupdate' => NULL,
      'qdelete' => NULL,
      'qlist'   => NULL,
      'qcount'  => NULL,
      'connection'   => [
        // A Drupal setting.php database key.
        'dbkey' => '',
        // An array of schema name strings or an empty array for default schema.
        'schemas' => [],
      ],
      // Structure: array of ['placeholder' => string placeholder name,
      // 'query' => string SQL query to return corresponding value.
      'placeholders' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    parent::setConfiguration($configuration);
    // Update placeholders.
    $this->initPlaceholders(TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    // Database and schemas selection.
    $form_state->setCached(FALSE);
    $db_settings_open = $form_state->get('db_settings_open') ??
      (
        !empty($this->configuration['connection']['schemas'][0])
        && ($this->configuration['connection']['schemas'][0] != static::DEFAULT_SCHEMA)
      );
    $form['connection'] = [
      '#type' => 'details',
      '#title' => $this->t('Connection settings'),
      '#open' => $db_settings_open,
    ];
    $form['connection']['dbkey'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Secondary database key name (optional)'),
      '#description' => $this->t(
        "Leave this field empty if you don't know how to use it or if you are
        not using another database than the Drupal one. Otherwise, you can use
        the database \"key\" used to specify another database in your Drupal site
        \"settings.php\" file."
      ),
      '#default_value' => $this->configuration['connection']['dbkey'],
    ];
    $form['connection']['description'] = [
      '#markup' =>
      '<div>'
      . $this->t(
        "You may specify one or more schemas if you don't want to use
        just the default one."
      )
      . '</div>',
    ];
    $schema_count = $form_state->get('schema_count');
    if (empty($schema_count)) {
      $schema_count = count($this->configuration['connection']['schemas']) + 1;
    }
    $form_state->set('schema_count', $schema_count);

    for ($i = 0; $i < $schema_count; ++$i) {
      $value = $this->configuration['connection']['schemas'][$i] ?? '';
      $form['connection']['schemas'][$i] = [
        '#type' => 'textfield',
        '#title' => $this->t('Schema name'),
        '#default_value' => $value,
      ];
    }
    $form['connection']['add_schema'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add a schema'),
      '#name' => 'add_schema',
      '#ajax' => [
        'callback' => [
          get_class($form_state->getFormObject()),
          'buildAjaxStorageClientConfigForm',
        ],
        'wrapper' => 'external-entities-storage-client-config-form',
        'method' => 'replace',
        'effect' => 'fade',
      ],
    ];

    // Interface for placeholder queries.
    // Fields: 'placeholder', 'query'.
    $placeholders_open = $form_state->get('placeholders_open') ?? FALSE;
    $form['placeholders'] = [
      '#type' => 'details',
      '#title' => $this->t('Placeholders'),
      '#open' => $placeholders_open,
    ];
    $form['placeholders']['description'] = [
      '#markup' =>
      '<div>'
      . $this->t(
        'You can define placeholders that will be replaced by cached values in
        your queries. Thoses values are fetched using the provided SQL query
        when the database external entity type is saved or when the cache
        needs to be updated. Therefore, you can use those placeholders in your
        SQL queries below and avoid running subqueries or additional joins to
        your original query in order to get a given value that relies on
        static data.'
      )
      . '</div><div>'
      . $this->t('For each placeholder, you must specify the name of the placeholder to use in queries (including leading ":"). It must start with ":" followed by a letter and only contain alphanumeric characters and underscores. You can leave it empty to remove a placeholder setting.')
      . '</div><div>'
      . $this->t('You must also specify a SQL query to use to fetch the value(s). The query must return only one column. Additionnal columns will be ignored. Multiple rows are supported but the placeholder should be suffixed with "[]" (eg.: ":multi_ids[]").')
      . '</div>'
      ,
    ];
    $placeholder_count = $form_state->get('placeholder_count');
    if (empty($placeholder_count)) {
      $placeholder_count = count($this->configuration['placeholders']) + 1;
    }
    $form_state->set('placeholder_count', $placeholder_count);

    for ($i = 0; $i < $placeholder_count; ++$i) {
      $value = $this->configuration['placeholders'][$i] ?? FALSE;
      // Try to get a sample value.
      $current_value = NULL;
      if (!empty($value['placeholder'])) {
        $current_value = $this->replacePlaceholders($value['placeholder']);
        // Check if it has been replaced.
        if ($current_value == $value['placeholder']) {
          $current_value = NULL;
        }
      }

      $form['placeholders']['placeholders'][$i] = [
        '#type' => 'fieldset',
        'placeholder' => [
          '#type' => 'textfield',
          '#title' => $this->t('Placeholder'),
          '#title_display' => 'before',
          '#default_value' => ($value ? $value['placeholder'] : ''),
        ],
        'query' => [
          '#type' => 'textarea',
          '#title' => $this->t('SQL query'),
          '#title_display' => 'before',
          '#default_value' => ($value ? $value['query'] : ''),
          '#description' => !empty($current_value) ? $this->t('Current value: @current_value', ['@current_value' => $current_value]) : '',
        ],
      ];
    }
    $form['placeholders']['add_placeholder'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add a placeholder'),
      '#name' => 'add_placeholder',
      '#ajax' => [
        'callback' => [
          get_class($form_state->getFormObject()),
          'buildAjaxStorageClientConfigForm',
        ],
        'wrapper' => 'external-entities-storage-client-config-form',
        'method' => 'replace',
        'effect' => 'fade',
      ],
    ];

    $form['queries'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('CRUD + List'),
      '#description' => $this->t(
        'CRUD + List SQL queries used to manage database objects. For each query
        type, you may use more than one SQL statement by separating them with
        ";". You may also use placeholders as defined above. You should specify
        tables using the notation "{#:tablename}" where "#" is a schema index
        starting from 1 and corresponding to the schema specified above (just
        use "1" if you don\'t know). You may also specify Drupal tables if
        needed using either the regular Drupal notation "{tablename}" or
        {0:tablename}.'
      ),
    ];
    $form['queries']['qcreate'] = [
      '#type' => 'textarea',
      '#title' => $this->t('CREATE: SQL queries to create a new object'),
      '#description' => $this->t(
        'Each query must contain the fields it suppors using placeholders
        ":<field_machine_name>". For instance, to insert a stock that has
        "name"="My Stock", "uniquename"="MYSTOCK001" and
        "type_name"="stock_types:my type" (using the format
        "cv.name:cvterm.name") fields mapped, use: "INSERT INTO {1:stock} (name,
        uniquename, type_id) SELECT :name, :uniquename, cvterm_id FROM
        {1:cvterm} cvt JOIN {1:cv} cv USING (cv_id) WHERE cvt.name =
        substr(:type_name, strpos(:type_name, \':\') + 1) AND cv.name =
        substr(:type_name, 1, strpos(:type_name, \':\') - 1) LIMIT 1;". The last
        query MUST return the created item identifier value aliased as "id" (ex.
        for PostgreSQL, it would be the last sequence value:
        "SELECT currval(\'&lt;name_id_seq&gt;\') AS id;" for MySQL it would be:
        "SELECT last_insert_id() AS id;").'
      ),
      '#required' => FALSE,
      '#default_value' => $this->configuration['qcreate'],
    ];

    $form['queries']['qread'] = [
      '#type' => 'textarea',
      '#title' => $this->t('READ: SQL queries to get a full object'),
      '#description' => $this->t(
        'The query can be complex with joins and will return the wanted fields
        named using "AS" clause (otherwise, the table field names will be used).
        Returned field names starting with "array_" will be parsed as database
        arrays and will be turned into array of values (ie. mapping fields
        should have a cardinality greater than 1). The query must contain a
        placeholder ":id" refering to the object identifier which field is
        aliased as "id". The ":id" placeholders used in queries must be used as
        right operands with either "= :id" or "IN (:id)" operators. Fields names
        should be distinct, contain only alphanumeric and underscore characters
        and must not start with a number.
        Ex.: "SELECT s.stock_id AS "id", s.name AS "name", s.uniquename AS
        "uniquename", cv.name || \':\' || cvt.name AS "type_name" FROM {1:stock}
        s JOIN {1:cvterm} cvt ON cvt.cvterm_id = s.type_id JOIN {1:cv} cv ON
        cv.cv_id = cvt.cv_id WHERE s.stock_id = :id;". When multiple queries are
        used, new field values are added to the entity and existing ones are
        replaced by the last queries.'
      ),
      '#required' => TRUE,
      '#default_value' => $this->configuration['qread'],
    ];

    $form['queries']['qupdate'] = [
      '#type' => 'textarea',
      '#title' => $this->t('UPDATE: SQL queries to update an existing object'),
      '#description' => $this->t(
        'Multiple queries are allowed separated by ";". Each query must contain
        the fields it supports using placeholders ":<field_machine_name>". For
        instance, to update a stock that has "name", "uniquename" and
        "type_name" (using the format "cv.name:cvterm.name") fields mapped, use:
        "UPDATE {1:stock} SET name = :name, uniquename = :uniquename, type_id =
        (SELECT :name, :uniquename, cvterm_id FROM {1:cvterm} cvt JOIN {1:cv} cv
        USING (cv_id) WHERE s.stock_id = :stock_id AND cvt.name =
        substr(:type_name, strpos(:type_name, \':\') + 1) AND cv.name =
        substr(:type_name, 1, strpos(:type_name, \':\') - 1) LIMIT 1);"'
      ),
      '#required' => FALSE,
      '#default_value' => $this->configuration['qupdate'],
    ];

    $form['queries']['qdelete'] = [
      '#type' => 'textarea',
      '#title' => $this->t('DELETE: SQL queries to delete an object'),
      '#description' => $this->t(
        'Each query must contain the fields it needs using placeholders
        ":<field_machine_name>". For instance, to delete a given stock, use:
        "DELETE FROM {1:stock} WHERE stock_id = :id;"'
      ),
      '#required' => FALSE,
      '#default_value' => $this->configuration['qdelete'],
    ];

    $form['queries']['qlist'] = [
      '#type' => 'textarea',
      '#title' => $this->t('LIST: SQL query to list objects'),
      '#description' => $this->t(
        'There must be only one statement for listing. The query must return
        object IDs and names and *MUST NOT* include "ORDER BY", "LIMIT" and
        "OFFSET" clauses which will be appended automatically to the query by
        the system. It should include a placeholder ":filters" in a WHERE clause
        to indicate where filters could be added by the system in order to
        filter the list.'
      ),
      '#required' => TRUE,
      '#default_value' => $this->configuration['qlist'],
    ];

    $form['queries']['filter_mappings'] = [
      '#type' => 'details',
      '#title' => $this->t('Filter mappings'),
      '#open' => FALSE,
      '#description' => $this->t('In order to allow filtering, Drupal field
        names need to be associated with SQL column names (or expressions).
        Filters based on unassociated fields will be ignored. Currently, this
        only works with simple fields (those with only one column, as opposed to
        compound fields, such as latitude/longitude). You may need to reload
        this page after saving field mapping changes.
      '),
    ];

    if (!empty($this->externalEntityType)
        && ($this->externalEntityType instanceof ExternalEntityTypeInterface)
    ) {
      $entity_type = $this->externalEntityType->getDerivedEntityTypeId();
      $field_definitions = $this
        ->entityFieldManager
        ->getFieldDefinitions($entity_type, $entity_type);
      foreach ($field_definitions as $field_machine_name => $field_definition) {
        if (strpos($field_machine_name, 'annotation') === 0) {
          // Annotations are not relevant.
          continue;
        }

        $form['queries']['filter_mappings']['mappings'][$field_machine_name] = [
          '#type' => 'textfield',
          '#title' => $this->t('@human_name (@machine_name)', [
            '@human_name' => $field_definition->getLabel(),
            '@machine_name' => $field_machine_name,
          ]),
          '#default_value' => $this->configuration['filter_mappings'][$field_machine_name] ?? '',
        ];
      }
    }

    $form['queries']['qcount'] = [
      '#type' => 'textarea',
      '#title' => $this->t('COUNT: SQL query to count objects'),
      '#description' => $this->t(
        'There must be only one statement for counting. The query must return
        just one integer as "count".'
      ),
      '#required' => TRUE,
      '#default_value' => $this->configuration['qcount'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * Note: there is no ::submitConfigurationForm because of the PluginForm
   * system. All is done here.
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    // Check for Ajax events.
    if ($trigger = $form_state->getTriggeringElement()) {
      switch ($trigger['#name']) {
        case 'add_schema':
          $schema_count = $form_state->get('schema_count');
          $form_state->set('schema_count', $schema_count + 1);
          $form_state->set('db_settings_open', TRUE);
          $form_state->setRebuild(TRUE);
          break;

        case 'add_placeholder':
          $placeholder_count = $form_state->get('placeholder_count');
          $form_state->set('placeholder_count', $placeholder_count + 1);
          $form_state->set('placeholders_open', TRUE);
          $form_state->setRebuild(TRUE);
          break;

        default:
      }
    }

    // @todo check query structure (single/multitple), presence of ':id', etc.
    $queries = $form_state->getValue('queries');
    $form_state->setValue('qcreate', $queries['qcreate'] ?? '');
    $form_state->setValue('qread', $queries['qread'] ?? '');
    $form_state->setValue('qupdate', $queries['qupdate'] ?? '');
    $form_state->setValue('qdelete', $queries['qdelete'] ?? '');
    $form_state->setValue('qlist', $queries['qlist'] ?? '');
    $form_state->setValue('qcount', $queries['qcount'] ?? '');
    $form_state->setValue('filter_mappings', $queries['filter_mappings']['mappings'] ?? []);

    $schemas = array_filter(
      ($form_state->getValue('connection') ?? ['schemas' => []]
      )['schemas']
    );
    $dbkey = ($form_state->getValue('connection') ?? ['dbkey' => ''])['dbkey'];
    $form_state->setValue('connection', [
      'dbkey' => $dbkey,
      'schemas' => $schemas,
    ]);

    // Only keep fully filled values.
    $placeholders = ($form_state->getValue('placeholders')
      ?? ['placeholders' => []])['placeholders'];
    $placeholders = array_filter($placeholders,
      function ($item) {
        return !(empty($item['placeholder']) || empty($item['query']));
      }
    );
    $form_state->setValue('placeholders', $placeholders);
    // @todo warn for invalid placeholder queries.
    // @todo make sure every ':placeholder' in queries has been defined.
    $this->setConfiguration($form_state->getValues());
  }

  /**
   * {@inheritdoc}
   */
  public function delete(ExternalEntityInterface $entity) {
    $entity_data = $entity->toRawData();
    $success = $this->runQueries($this->configuration['qdelete'], $entity_data);
    if (empty($success)) {
      $this->messenger->addError(
        'Failed to delete entity ' . $entity->id()
      );
    }
    return SAVED_DELETED;
  }

  /**
   * Expands a database array of a field value.
   *
   * Database driver used returns array field results as strings starting with
   * '{' and ending with '}' which values are separated by comas ",".
   *
   * @param string $sql_array
   *   A field value containing a database array.
   *
   * @return array
   *   The array of values.
   */
  protected function expandDatabaseArray($sql_array) :array {
    // @todo support ',' inside values which will cause invalid string
    // explosion.
    // Removes leading '{' and trailing '}' and explode string on ','.
    return explode(',', substr($sql_array, 1, -1));
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(array $ids = NULL) {
    $data = [];

    try {
      // Replaces "=:id" or "IN(:id)" with "IN(:id[])".
      // Note: replacing !=:id, >=:id and <=:id is not supported.
      $queries = preg_replace(
        ['/(?<![!<>])=\s*:id(\W|$)/', '/(\W)IN\s*\(:id(?:\[\])\)?/i'],
        [' IN (:id[])\1', '\1IN (:id[])'],
        $this->configuration['qread']
      );
      // @todo make sure $queries does not contain any ':id' left (not followed
      // by '[]').
      $queries = $this->replacePlaceholders($queries);
      $sql_queries = explode(';', $queries);
      // Loop on queries.
      foreach ($sql_queries as $sql_query) {
        $sql_query = trim($sql_query);
        if (empty($sql_query)) {
          continue;
        }
        // @todo find a way to get the field name mapped to "id" and use that
        // name instead of 'id'.
        $results = $this->xConnection->query(
          $sql_query,
          [':id[]' => $ids]
        )->fetchAllAssoc('id', \PDO::FETCH_ASSOC);

        // Restructures PostgreSQL arrays.
        foreach ($results as $id => $element) {
          foreach ($element as $key => $value) {
            if (0 === strpos($key, 'array_')) {
              $data[$id][$key] = $this->expandDatabaseArray($value);
            }
            else {
              $data[$id][$key] = $value;
            }
          }
        }
      }
    }
    catch (DatabaseExceptionWrapper $e) {
      $this->messenger->addError('Query failed: ' . $e->getMessage());
      $this->logger->error($e);
    }
    if (empty($data)) {
      $data = [];
    }
    return $data;
  }

  /**
   * Loads one entity.
   *
   * @param mixed $id
   *   The ID of the entity to load.
   *
   * @return array|null
   *   A raw data array, NULL if no data returned.
   */
  public function load($id) {
    $entity = $this->runQueries($this->configuration['qread'], ['id' => $id]);
    if (empty($entity)) {
      $this->messenger->addError('Failed to load entity ' . $id);
      $entity = NULL;
    }
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function save(ExternalEntityInterface $entity) {
    // Check if ID exists in base.
    $entity_data = $entity->toRawData();
    if ($entity->id()) {
      $updated_entity = $this->runQueries(
        $this->configuration['qupdate'], $entity_data
      );
      if (empty($updated_entity)) {
        $this->messenger->addError(
          'Failed to update entity ' . $entity->id()
        );
      }
      else {
        $result = SAVED_UPDATED;
      }
    }
    else {
      $new_entity = $this->runQueries(
        $this->configuration['qcreate'], $entity_data
      );
      if (empty($new_entity)) {
        $this->messenger->addError('Failed to create new entity.');
      }
      else {
        $result = $new_entity['id'] ?? SAVED_NEW;
      }
    }
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function query(
    array $parameters = [],
    array $sorts = [],
    $start = NULL,
    $length = NULL
  ) {
    try {
      $start = $start ?? 0;
      $query = rtrim($this->configuration['qlist'], ';')
        . ($sorts ? ' ORDER BY ' . implode(',', array_map(function ($x) {
            return $x['field'] . ' ' . $x['direction'];
        }, $sorts)) : '')
        . ($length ? ' LIMIT ' . $length : '')
        . ($start ? ' OFFSET ' . $start : '');
      $query = $this->replacePlaceholders($query, [':filters' => $this->translateParameters($parameters)]);
      $results = $this->xConnection->query(
        $query
      )->fetchAll(\PDO::FETCH_ASSOC);
    }
    catch (DatabaseExceptionWrapper $e) {
      $this->messenger->addError('Query failed: ' . $e->getMessage());
      $this->logger->error($e);
    }

    return $results;
  }

  /**
   * Translates the string operators to SQL equivalents.
   *
   * @param array $parameters
   *   An array of condition arrays.
   *
   * @note this only works for fields whose placeholder names directly match SQL
   * column names.
   */
  protected function translateParameters(array $parameters) {
    $conditions = [];
    // Dummy table in order to get a SelectInterface object.
    $s = $this->xConnection->select('-DUMMY-');

    foreach ($parameters as &$parameter) {
      // Replace entity field name with associated SQL column name or
      // expression.
      if ($real_name = $this->getFilterMapping($parameter['field'])) {
        $parameter['field'] = $real_name;
      }
      else {
        // This cannot be guessed.
        continue;
      }

      // Translate non-SQL conditions such as CONTAINS.
      Condition::translateCondition($parameter, $s, FALSE);

      // Build SQL condition strings.
      if (is_string($parameter['value'])) {
        $parameter['value'] = "'" . $parameter['value'] . "'";
      }
      if ($parameter['operator'] === 'LIKE') {
        // Deal with capitalization.
        $parameter['operator'] = 'ILIKE';
      }
      if ($parameter['operator'] == 'IN' && is_array($parameter['value'])) {
        $parameter['value'] = '(' . implode(', ', $parameter['value']) . ')';
      }
      $conditions[] = $parameter['field'] . ' ' . $parameter['operator'] . ' ' . $parameter['value'];
    }

    return empty($conditions) ? '' : 'AND ' . implode(' AND ', $conditions);
  }

  /**
   * Get SQL column name or expression associated with a field.
   *
   * @param string $field_machine_name
   *   Field machine name.
   *
   * @return string
   *   SQL column name or expression.
   */
  protected function getFilterMapping($field_machine_name) {
    return $this->configuration['filter_mappings'][$field_machine_name] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function countQuery(array $parameters = []) {
    // @todo manage parameters for filtering (array of ['field' => $fn,
    // 'value' => $val, 'operator' => $op]).
    // See ExternalEntityStorageClientInterface::countQuery
    $count = 0;
    try {
      $query = rtrim($this->configuration['qcount'], ';');
      $query = $this->replacePlaceholders($query);
      $count = $this->xConnection->query($query)->fetch()->count;
    }
    catch (DatabaseExceptionWrapper $e) {
      $this->messenger->addError('Query failed: ' . $e->getMessage());
      $this->logger->error($e);
    }

    return $count;
  }

  /**
   * Handles query placeholders and perform a set of SQL queries.
   *
   * @param string $queries
   *   A set of queries separated by ';'.
   * @param array $entity_data
   *   Entity fields as an associative array.
   *
   * @return array
   *   An empty array if a query failed and non-empty array otherwise. The array
   *   will contain an aggregation of query results keyed by field names. In
   *   case of multiple queries returning the same field name, only the last
   *   value will be kept. If no fields are returned, the array will just
   *   contain the "TRUE" value.
   */
  protected function runQueries(string $queries, array $entity_data) :array {
    // Start a new transaction.
    $transaction = $this->xConnection->startTransaction();
    $return = [];
    try {
      // @todo better parse ';' as it can be used in query strings.
      $sql_queries = explode(';', $queries);
      // Loop on queries.
      foreach ($sql_queries as $sql_query) {
        $sql_query = trim($sql_query);
        if (empty($sql_query)) {
          continue;
        }
        $values = [];
        // Get the list of field values needed.
        preg_match_all(
          '/\W:([a-z_]\w*(?:\[\])?)/i',
          $sql_query,
          $placeholders,
          PREG_PATTERN_ORDER
        );
        $placeholders = array_flip($placeholders[1]);
        foreach ($placeholders as $placeholder => $waste) {
          // @todo manage array values, str_contains($placeholder, '[]').
          if (array_key_exists($placeholder, $entity_data)) {
            $values[":$placeholder"] = $entity_data[$placeholder];
          }
        }
        // Replace placeholders.
        $sql_query = $this->replacePlaceholders($sql_query);
        // @todo make sure all placeholders are managed or raise an error.
        $result = $this->xConnection->query(
          $sql_query,
          $values
        )->fetch(\PDO::FETCH_ASSOC);
        foreach ($result as $key => $value) {
          if (0 === strpos($key, 'array_')) {
            // Removes leading '{' and trailing '}' and explode string on ','.
            // @todo use a protected method and support ',' inside values which
            // will cause invalid string explosion.
            $return[$key] = $this->expandDatabaseArray($value);
          }
          else {
            $return[$key] = $value;
          }
        }
      }
      // Make sure we return a non-empty array on success.
      if (empty($return)) {
        $return = [TRUE];
      }
    }
    catch (DatabaseExceptionWrapper $e) {
      $transaction->rollBack();
      $this->messenger->addError('Query failed: ' . $e->getMessage());
      $this->logger->error($e);
    }
    // Commit.
    unset($transaction);
    return $return;
  }

  /**
   * Initializes placeholder replacement.
   *
   * @param bool $reload_cache
   *   If TRUE, it reloads identifier cache.
   *
   * @return array
   *  Returns the list of placeholders keyed by external entity type, then by
   *  placeholder name with their associated values.
   */
  public function initPlaceholders(bool $reload_cache = FALSE) :array {
    // Get cache.
    $cid = 'xnttdb:placholders';
    $placholders = [];
    if ($cache = $this->cache->get($cid)) {
      $placholders = $cache->data;
    }

    // Checks if the external entity type's mapping has already been processed.
    if ($this->externalEntityType) {
      $entity_type = $this->externalEntityType->getDerivedEntityTypeId();
      if (!isset($placholders[$entity_type])) {
        $reload_cache = TRUE;
      }
    }
    else {
      $entity_type = '';
      $reload_cache = TRUE;
    }

    // Load placeholder replacement values.
    if ($reload_cache && $this->xConnection) {
      $placeholders_issues = FALSE;
      $placholders[$entity_type] = [];
      // Process the list of placeholders from config.
      foreach ($this->configuration['placeholders'] as $ph_def) {
        // Skip placeholders not starting with ':'.
        if (!preg_match('/^:[a-z]\w*(?:\[\])?$/i', $ph_def['placeholder'])) {
          $placeholders_issues = TRUE;
          $this->logger->warning('Ignoring invalid placeholder name "' . $ph_def['placeholder'] . '"');
          continue;
        }
        // Process placeholder query.
        try {
          $results = $this->xConnection->query(
            $ph_def['query']
          )->fetchAll(\PDO::FETCH_ASSOC);
          if (!empty($results)) {
            // Save placeholder replacement value.
            if (str_ends_with($ph_def['placeholder'], '[]')) {
              $placholders[$entity_type][$ph_def['placeholder']] = array_map(
                function ($x) {
                  return $x[array_key_first($x)];
                },
                $results
              );
            }
            else {
              $placholders[$entity_type][$ph_def['placeholder']] =
                $results[0][array_key_first($results[0])];
            }
          }
          else {
            $this->messenger->addWarning(
              'Failed to get value(s) for placeholder "' . $ph_def['placeholder'] . '". Query "' . $ph_def['query'] . '" returned no value.'
            );
          }
        }
        catch (DatabaseExceptionWrapper $e) {
          $this->messenger->addWarning(
            'Failed to get value(s) for placeholder "' . $ph_def['placeholder'] . '" (database error). '
            . $e->getMessage()
          );
        }
        catch (\Throwable $e) {
          $this->messenger->addWarning(
            'Failed to get value(s) for placeholder "' . $ph_def['placeholder'] . '". '
            . $e->getMessage()
          );
        }
      }
      if ($placeholders_issues) {
        $this->messenger->addWarning(
          'A couple of placeholder names were invalid and ignored. Please check the external entity configuration and see logs for details.'
        );
      }

      $this->cache->set($cid, $placholders);
    }

    return $this->placeholders = $placholders;
  }

  /**
   * Replace placeholders in a query.
   *
   * @param string $query
   *   An SQL query containing placeholders of the form ':placeholdername'.
   * @param array $additional_replacements
   *   Additional replacement patterns, such as :filters.
   *
   * @return string
   *   The query containing actual replacement values.
   */
  public function replacePlaceholders(string $query, array $additional_replacements = []) :string {
    $entity_type = '';
    if ($this->externalEntityType) {
      $entity_type = $this->externalEntityType->getDerivedEntityTypeId();
    }

    // Replaces placeholders.
    $this->placeholders[$entity_type] = $this->placeholders[$entity_type] ?? [];
    $placeholders = $this->placeholders[$entity_type] + $additional_replacements;
    array_walk(
      $placeholders,
      function (&$item, $key) {
        if (is_array($item)) {
          $item = implode(', ', $item);
        }
      }
    );
    // Reverse sort by key to replace longuest placeholders first and avoid
    // having a part of a placeholder replaced by another.
    // ex.: ':abc' with value 42 and ':abcdef' with value 806. If we replace
    // first ':abc', a placeholder ':abcdef' would be turned into "42def"
    // instead of "806".
    krsort($placeholders);
    $query = str_replace(
      array_keys($placeholders),
      array_values($placeholders),
      $query
    );
    return $query;
  }

}
